from typing import Optional

from PIL import Image

from ..codec import ImageCodecAdaptor, supportsFormats
from ..constants import CompressionFormat, CompressionQuality
from ..exceptions import DependancyMissingException
from ..lib.imageOps import switchRGBA_BGRA

try:
    import tex2img
except ImportError as e:
    raise DependancyMissingException("Tex2ImgCodec", "tex2img", e)

TEX2IMG_FORMAT_MAP = {
    # CompressionFormat.PVRTC14A: 12, # Tex2ImgCodec segfaults decoding PVRTC with all tests (#29)
    CompressionFormat.ATC: 14, #cATC_RGBA_INTERPOLATED_ALPHA
    CompressionFormat.DXT1: 5,
    CompressionFormat.DXT5: 6,
    CompressionFormat.ETC1: 0,
    CompressionFormat.ETC2: 2
}

# tex2img seems to swap ETC2's R and B channels - but not ETC1?
SWAP_CHANNELS_POST = { CompressionFormat.ETC2 }

@supportsFormats(decompresses=TEX2IMG_FORMAT_MAP.keys())
class Tex2ImgCodec(ImageCodecAdaptor):
    @classmethod
    def decompress(cls, fp: bytes, format: CompressionFormat, width: int, height: int, quality: Optional[CompressionQuality]) -> Image.Image:
        if format not in TEX2IMG_FORMAT_MAP:
            raise ValueError(f"Codec {Tex2ImgCodec.__name__} does not support format {format.name}")
        
        decompressed = tex2img.basisu_decompress(fp, width, height, TEX2IMG_FORMAT_MAP[format]) # type: ignore[reportUnknownMemberType]
        im = Image.frombytes("RGBA", (width, height), decompressed, "raw") # type: ignore[reportUnknownMemberType]
        
        if format in SWAP_CHANNELS_POST:
            with im:
                im = switchRGBA_BGRA(im)
        
        return im



from PIL import Image

def switchRGBA_BGRA(im: Image.Image):
    """Swap the red and blue channels of an image, and return as a new image.
    This method does not discern between RGB/BGR images.

    :param im: The image whose channels to swap
    :type im: Image
    :return: `im`, but with the red and blue channels swapped
    :rtype: Image
    """
    if im.mode == "RGB":
        r, g, b = im.split()
        return Image.merge("RGB", (b, g, r))
    
    elif im.mode == "RGBA":
        r, g, b, a = im.split()
        return Image.merge("RGBA", (b, g, r, a))
    
    raise ValueError("Only RGB/RGBA images are accepted")



from typing import Tuple


class Texture:
    def __init__(self, x: int, y: int, width: int, height: int) -> None:
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    
    @property
    def shape(self) -> Tuple[int, int]:
        """The shape of the texture region, given as a (width, height) tuple.

        :return: The shape of the texture region, as a (width, height) tuple
        :rtype: Tuple[int, int]
        """
        return (self.width, self.height)
    

    @shape.setter
    def shape(self, value: Tuple[int, int]) -> None:
        """Set the shape of the texture region.

        :param value: The shape of the texture region, as a (width, height) tuple
        :type value: Tuple[int, int]
        """
        (self.width, self.height) = value

    
    @property
    def position(self) -> Tuple[int, int]:
        """The coordinates of the texture region relative to the AEI origin, given as an (x, y) tuple.

        :return: The coordinates of the texture region, as an (x, y) tuple
        :rtype: Tuple[int, int]
        """
        return (self.x, self.y)

    
    @position.setter
    def position(self, value: Tuple[int, int]) -> None:
        """Set the coordinates of the texture region relative to the AEI origin.

        :param value: The coordinates of the texture region, as an (x, y) tuple
        :type value: Tuple[int, int]
        """
        (self.x, self.y) = value

    def __str__(self):
        return f"Texture: x: {self.x}, y: {self.y}, w: {self.width}, h: {self.height}"
    
    def equals(self, other: "Texture") -> bool: 
        if not isinstance(other, Texture):
            return False
        return  (
            self.x == other.x and
            self.y == other.y and
            self.width == other.width and
            self.height == other.height)


from typing import Optional

from PIL.Image import Image
from contextlib import nullcontext

from ..codec import ImageCodecAdaptor, supportsFormats
from ..constants import CompressionFormat, CompressionQuality
from ..exceptions import DependancyMissingException

try:
    import etcpak
except ImportError as e:
    raise DependancyMissingException("EtcPakCodec", "etcpak", e)


@supportsFormats(compresses=[
    CompressionFormat.DXT5,
    CompressionFormat.ETC1,
    CompressionFormat.ETC2
])
class EtcPakCodec(ImageCodecAdaptor):
    @classmethod
    def compress(cls, im: Image, format: CompressionFormat, quality: Optional[CompressionQuality]) -> bytes:
        imageIn, ctx = cls._ensureRgba(im)
        with ctx:
            if format is CompressionFormat.DXT5:
                return etcpak.compress_to_dxt5(imageIn.tobytes(), imageIn.width, imageIn.height) # type: ignore[reportUnknownVariableType]

            elif format is CompressionFormat.ETC1:
                return etcpak.compress_to_etc1(imageIn.tobytes(), imageIn.width, imageIn.height) # type: ignore[reportUnknownVariableType]
            
            elif format is CompressionFormat.ETC2:
                # etcpak does provide a function for etc2 compression, but it produces almost completely black images
                # ETC2 is backwards compatible with ETC1, so as a stopgap we'll just compress as etc1
                # https://www.khronos.org/assets/uplo...pengl-es-bof/Ericsson-ETC2-SIGGRAPH_Aug12.pdf
                return etcpak.compress_to_etc1(imageIn.tobytes(), imageIn.width, imageIn.height) # type: ignore[reportUnknownVariableType]
            
        raise ValueError(f"Codec {EtcPakCodec.__name__} does not support format {format.name}")
    
    
    @classmethod
    def _ensureRgba(cls, im: Image):
        converted = im.mode != "RGBA"
        if converted:
            rgbaImg = im.convert("RGBA")
        else:
            rgbaImg = im
        
        return rgbaImg, (rgbaImg if converted else nullcontext())
